"""Transform a ``CompilerISA`` to an ``nx.Graph``."""

import networkx as nx

from pyquil.external.rpcq import CompilerISA


def compiler_isa_to_graph(compiler_isa: CompilerISA) -> nx.Graph:
    """Generate an ``nx.Graph`` based on the qubits and edges of any ``CompilerISA``."""
    return nx.from_edgelist([int(i) for i in edge.ids] for edge in compiler_isa.edges.values())



"""Classes that represent the control flow graph of a Quil program."""

from typing import Optional

from quil import program as quil_rs
from typing_extensions import Self, override

from pyquil.quilbase import (
    AbstractInstruction,
    _convert_to_py_instruction,
    _convert_to_py_instructions,
)


class BasicBlock(quil_rs.BasicBlock):
    """Represents a basic block in the Program.

    Most functionality is implemented by the `quil` package. See the
    `quil BasicBlock documentation`_ for documentation and available methods.

    .. _quil BasicBlock documentation: https://rigetti.github.io/quil-rs/quil/program.html#BasicBlock
    """

    @classmethod
    def _from_rs(cls, block: quil_rs.BasicBlock) -> Self:
        return super().__new__(cls, block)

    @override
    def instructions(self) -> list[AbstractInstruction]:  # type: ignore[override]
        return _convert_to_py_instructions(super().instructions())

    @override
    def terminator(self) -> Optional[AbstractInstruction]:  # type: ignore[override]
        inst = super().terminator()
        if inst is None:
            return None
        return _convert_to_py_instruction(super().terminator())


class ControlFlowGraph(quil_rs.ControlFlowGraph):
    """Representation of a control flow graph (CFG) for a Quil program.

    The CFG is a directed graph where each node is a basic block and each edge is a control flow transition between two
    basic blocks.

    This class should not be initialized directly. Use :py:meth:~pyquil.quil.Program.control
     flow_graph` to get a CFG for a program.

    Most functionality is implemented by the `quil` package. See the `quil ControlFlowGraph documentation`_ for
    available methods.

    .. _quil ControlFlowGraph documentation: https://rigetti.github.io/quil-rs/quil/program.html#ControlFlowGraph
    """

    @classmethod
    def _from_rs(cls, graph: quil_rs.ControlFlowGraph) -> Self:
        return super().__new__(cls, graph)

    @override
    def basic_blocks(self) -> list[BasicBlock]:  # type: ignore[override]
        """Return a list of all the basic blocks in the control flow graph, in order of definition."""
        return [BasicBlock._from_rs(block) for block in super().basic_blocks()]



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from typing import Optional, Union, cast

import numpy as np
from qcs_sdk import QCSClient, qvm
from qcs_sdk.qvm import QVMOptions

from pyquil.api import MemoryMap
from pyquil.api._qvm import (
    validate_noise_probabilities,
)
from pyquil.gates import MOVE
from pyquil.paulis import PauliSum, PauliTerm
from pyquil.quil import Program
from pyquil.quilatom import MemoryReference
from pyquil.wavefunction import Wavefunction


class WavefunctionSimulator:
    def __init__(
        self,
        *,
        gate_noise: Optional[tuple[float, float, float]] = None,
        measurement_noise: Optional[tuple[float, float, float]] = None,
        random_seed: Optional[int] = None,
        timeout: float = 10.0,
        client_configuration: Optional[QCSClient] = None,
    ) -> None:
        """Return a simulator that propagates a wavefunction representation of a quantum state.

        :param gate_noise: A tuple of three numbers [Px, Py, Pz] indicating the probability of an X,
            Y, or Z gate getting applied to each qubit after a gate application or reset.
        :param measurement_noise: A tuple of three numbers [Px, Py, Pz] indicating the probability
            of an X, Y, or Z gate getting applied before a measurement.
        :param random_seed: A seed for the simulator's random number generators. Either None (for
            an automatically generated seed) or a non-negative integer.
        :param timeout: Time limit for requests, in seconds.
        :param client_configuration: Optional client configuration. If none is provided, a default one will be loaded.
        """
        validate_noise_probabilities(gate_noise)
        validate_noise_probabilities(measurement_noise)
        self.gate_noise = gate_noise
        self.measurement_noise = measurement_noise

        if random_seed is None:
            self.random_seed = None
        elif isinstance(random_seed, int) and random_seed >= 0:
            self.random_seed = random_seed
        else:
            raise TypeError("random_seed should be None or a non-negative int")

        self.timeout = timeout
        self._client = client_configuration or QCSClient.load()
        self._qvm_client = qvm.QVMClient.new_http(self._client.qvm_url)

    def wavefunction(self, quil_program: Program, memory_map: Optional[MemoryMap] = None) -> Wavefunction:
        """Simulate a Quil program and return the wavefunction.

        .. note:: If your program contains measurements or noisy gates, this method may not do what
            you want. If the execution of ``quil_program`` is **non-deterministic** then the
            final wavefunction only represents a stochastically generated sample and the
            wavefunctions returned by *different* ``wavefunction`` calls *will generally be
            different*.

        :param quil_program: A Quil program.
        :param memory_map: An assignment of classical registers to values, representing an initial
                           state for the QAM's classical memory.

                           This is expected to be of type dict[str, list[Union[int, float]]],
                           where the keys are memory region names and the values are arrays of
                           initialization data.
        :return: A Wavefunction object representing the state of the QVM.
        """
        if memory_map is not None:
            quil_program = self.augment_program_with_memory_values(quil_program, memory_map)

        request = qvm.api.WavefunctionRequest(
            quil_program.out(calibrations=False),
            self.measurement_noise,
            self.gate_noise,
            self.random_seed,
        )
        wavefunction = bytes(
            qvm.api.get_wavefunction(request, self._qvm_client, options=QVMOptions(timeout_seconds=self.timeout))
        )
        return Wavefunction.from_bit_packed_string(wavefunction)

    def expectation(
        self,
        prep_prog: Program,
        pauli_terms: Union[PauliSum, list[PauliTerm]],
        memory_map: Optional[dict[str, list[Union[int, float]]]] = None,
    ) -> Union[float, np.ndarray]:
        """Calculate the expectation value of Pauli operators given a state prepared by prep_program.

        If ``pauli_terms`` is a ``PauliSum`` then the returned value is a single ``float``,
        otherwise the returned value is an array of values, one for each ``PauliTerm`` in the
        list.

        .. note:: If your program contains measurements or noisy gates, this method may not do what
            you want. If the execution of ``quil_program`` is **non-deterministic** then the
            final wavefunction from which the expectation value is calculated only represents
            a stochastically generated sample and the wavefunctions returned by *different*
            ``wavefunction`` calls *will generally be different*.

        :param prep_prog: A program that prepares the state on which we measure the expectation.
        :param pauli_terms: A Pauli representation of a quantum operator.
        :param memory_map: An assignment of classical registers to values, representing an initial
                           state for the QAM's classical memory.

                           This is expected to be of type dict[str, list[Union[int, float]]],
                           where the keys are memory region names and the values are arrays of
                           initialization data.
        :return: Either a float or array floats depending on ``pauli_terms``.
        """
        is_pauli_sum = False
        if isinstance(pauli_terms, PauliSum):
            progs, coeffs = pauli_terms.get_programs()
            is_pauli_sum = True
        else:
            coeffs = np.array([pt.coefficient for pt in pauli_terms])
            progs = [pt.program for pt in pauli_terms]

        if memory_map is not None:
            prep_prog = self.augment_program_with_memory_values(prep_prog, memory_map)

        request = qvm.api.ExpectationRequest(prep_prog.out(), [prog.out() for prog in progs])
        expectations = qvm.api.measure_expectation(
            request, self._qvm_client, options=QVMOptions(timeout_seconds=self.timeout)
        )
        bare_results = np.asarray(expectations)
        results = coeffs * bare_results
        if is_pauli_sum:
            return np.sum(results)  # type: ignore
        return results  # type: ignore

    def run_and_measure(
        self,
        quil_program: Program,
        qubits: Optional[list[int]] = None,
        trials: int = 1,
        memory_map: Optional[MemoryMap] = None,
    ) -> np.ndarray:
        """Run a Quil program once to determine the final wavefunction, and measure multiple times.

        Alternatively, consider using ``wavefunction`` and calling ``sample_bitstrings`` on the
        resulting object.

        For a large wavefunction and a low-medium number of trials, use this function.
        On the other hand, if you're sampling a small system many times you might want to
        use ``Wavefunction.sample_bitstrings``.

        .. note:: If your program contains measurements or noisy gates, this method may not do what
            you want. If the execution of ``quil_program`` is **non-deterministic** then the
            final wavefunction from which the returned bitstrings are sampled itself only
            represents a stochastically generated sample and the outcomes sampled from
            *different* ``run_and_measure`` calls *generally sample different bitstring
            distributions*.

        :param quil_program: The program to run and measure
        :param qubits: An optional list of qubits to measure. The order of this list is
            respected in the returned bitstrings. If not provided, all qubits used in
            the program will be measured and returned in their sorted order.
        :param int trials: Number of times to sample from the prepared wavefunction.
        :param memory_map: An assignment of classical registers to values, representing an initial
                           state for the QAM's classical memory.

                           This is expected to be of type dict[str, list[Union[int, float]]],
                           where the keys are memory region names and the values are arrays of
                           initialization data.
        :return: An array of measurement results (0 or 1) of shape (trials, len(qubits))
        """
        if qubits is None:
            qubits = sorted(cast(set[int], quil_program.get_qubits(indices=True)))

        if memory_map is not None:
            quil_program = self.augment_program_with_memory_values(quil_program, memory_map)

        request = qvm.api.MultishotMeasureRequest(
            quil_program.out(),
            trials,
            qubits,
        )
        measured_qubits = qvm.api.run_and_measure(
            request, client=self._qvm_client, options=QVMOptions(timeout_seconds=self.timeout)
        )
        return np.asarray(measured_qubits)

    @staticmethod
    def augment_program_with_memory_values(
        quil_program: Program,
        memory_map: MemoryMap,
    ) -> Program:
        p = Program()

        # we stupidly allowed memory_map to be of type dict[MemoryReference, Any], whereas qc.run
        # takes a memory initialization argument of type dict[str, list[Union[int, float]]. until
        # we are in a position to remove this, we support both styles of input.

        if len(memory_map.keys()) == 0:
            return quil_program
        elif isinstance(list(memory_map.keys())[0], str):
            for name, arr in memory_map.items():
                for index, value in enumerate(arr):
                    p += MOVE(MemoryReference(name, offset=index), value)
        else:
            raise TypeError("Bad memory_map type; expected dict[str, list[Union[int, float]]].")

        p += quil_program

        return p



##############################################################################
# Copyright 2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import os
import shutil
import subprocess
import tempfile
from typing import Any, Optional

from IPython.display import Image

from pyquil.latex._diagram import DiagramSettings
from pyquil.latex._main import to_latex
from pyquil.quil import Program


def display(circuit: Program, settings: Optional[DiagramSettings] = None, **image_options: Any) -> Image:
    """Display a PyQuil circuit as an IPython image object.

    .. note::

       For this to work, you need two external programs, ``pdflatex`` and ``convert``,
       to be installed and accessible via your shell path.

       Further, your LaTeX installation should include class and style files for ``standalone``,
       ``geometry``, ``tikz``, and ``quantikz``. If it does not, you need to install
       these yourself.

    :param circuit: The circuit to be drawn, represented as a pyquil program.
    :param settings: An optional object of settings controlling diagram rendering and layout.
    :return: PNG image render of the circuit.
    """
    # The conversion process relies on two passes: first, 'pdflatex' is called to
    # render the tikz to a pdf. Second, Imagemagick's 'convert' is called to translate
    # this to a png image.
    pdflatex_path = shutil.which("pdflatex")
    convert_path = shutil.which("convert")

    if pdflatex_path is None:
        raise FileNotFoundError("Unable to locate 'pdflatex'.")
    if convert_path is None:
        raise FileNotFoundError("Unable to locate 'convert'.")

    with tempfile.TemporaryDirectory() as tmpdir:
        with open(os.path.join(tmpdir, "diagram.tex"), "w") as texfile:
            texfile.write(to_latex(circuit, settings))

        result = subprocess.run(
            [pdflatex_path, "-halt-on-error", "-output-directory", tmpdir, texfile.name],  # noqa: S603 - valid input
            stdout=subprocess.PIPE,
        )
        if result.returncode != 0:
            msg = f"'pdflatex' terminated with return code {result.returncode}."
            # NOTE: pdflatex writes all error messages to stdout
            if result.stdout:
                msg += " Transcript:\n\n" + result.stdout.decode("utf-8")
            raise RuntimeError(msg)

        png = os.path.join(tmpdir, "diagram.png")
        pdf = os.path.join(tmpdir, "diagram.pdf")

        result = subprocess.run([convert_path, "-density", "300", pdf, png], stderr=subprocess.PIPE)  # noqa: S603
        if result.returncode != 0:
            msg = f"'convert' terminated with return code {result.returncode}."
            if result.stderr:
                msg += " Transcript:\n\n" + result.stderr.decode("utf-8")
            raise RuntimeError(msg)

        return Image(filename=png, **image_options)



##############################################################################
# Copyright 2016-2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import json
from dataclasses import dataclass
from typing import Optional

from qcs_sdk import QCSClient
from qcs_sdk.compiler.quilc import (
    CompilerOpts,
    ConjugateByCliffordRequest,
    ConjugatePauliByCliffordResponse,
    GenerateRandomizedBenchmarkingSequenceResponse,
    NativeQuilMetadata,
    QuilcClient,
    RandomizedBenchmarkingRequest,
    TargetDevice,
    compile_program,
    conjugate_pauli_by_clifford,
    generate_randomized_benchmarking_sequence,
    get_version_info,
)
from rpcq.messages import TargetDevice as TargetQuantumProcessor


@dataclass
class CompileToNativeQuilRequest:
    """Request to compile to native Quil."""

    program: str
    """Program to compile."""

    target_quantum_processor: TargetQuantumProcessor
    """Quantum processor to target."""

    protoquil: Optional[bool]
    """Whether or not to restrict to protoquil. Overrides server default when provided."""


@dataclass
class NativeQuilMetadataResponse:
    """Metadata for a native Quil program."""

    final_rewiring: list[int]
    """Output qubit index relabeling due to SWAP insertion."""

    gate_depth: Optional[int]
    """Maximum number of successive gates in the native Quil program."""

    gate_volume: Optional[int]
    """Total number of gates in the native Quil program."""

    multiqubit_gate_depth: Optional[int]
    """Maximum number of successive two-qubit gates in the native Quil program."""

    program_duration: Optional[float]
    """Rough estimate of native Quil program length in nanoseconds."""

    program_fidelity: Optional[float]
    """Rough estimate of the fidelity of the full native Quil program."""

    topological_swaps: Optional[int]
    """Total number of SWAPs in the native Quil program."""

    qpu_runtime_estimation: Optional[float]
    """
    The estimated runtime (milliseconds) on a Rigetti QPU (protoquil program). Available only for protoquil-compliant
    programs.
    """


@dataclass
class CompileToNativeQuilResponse:
    """Compile to native Quil response."""

    native_program: str
    """Native Quil program."""

    metadata: Optional[NativeQuilMetadata]
    """Metadata for the returned Native Quil."""


class CompilerClient:
    """Client for making requests to a Quil compiler."""

    def __init__(
        self,
        *,
        client_configuration: QCSClient,
        request_timeout: float = 10.0,
        quilc_client: Optional[QuilcClient] = None,
    ) -> None:
        """Instantiate a new compiler client.

        :param client_configuration: Configuration for client.
        :param request_timeout: Timeout for requests, in seconds.
        """
        self._client_configuration = client_configuration
        base_url = client_configuration.quilc_url
        if not base_url.startswith("tcp://"):
            raise ValueError(f"Expected compiler URL '{base_url}' to start with 'tcp://'")

        if quilc_client is None:
            self.quilc_client = QuilcClient.new_rpcq(base_url)
        elif isinstance(quilc_client, QuilcClient):
            self.quilc_client = quilc_client
        else:
            raise TypeError(f"Unsupported type for Quilc client: {quilc_client}")

        self.base_url = base_url
        self.timeout = request_timeout

    def get_version(self) -> str:
        """Get version info for compiler server."""
        return get_version_info(client=self.quilc_client)

    def compile_to_native_quil(self, request: CompileToNativeQuilRequest) -> CompileToNativeQuilResponse:
        """Compile Quil program to native Quil."""
        target_device_json = json.dumps(request.target_quantum_processor.asdict())  # type: ignore
        target_device = TargetDevice.from_json(target_device_json)

        result = compile_program(
            quil=request.program,
            target=target_device,
            client=self.quilc_client,
            options=CompilerOpts(protoquil=request.protoquil, timeout=self.timeout),
        )
        return CompileToNativeQuilResponse(native_program=result.program, metadata=result.native_quil_metadata)

    def conjugate_pauli_by_clifford(self, request: ConjugateByCliffordRequest) -> ConjugatePauliByCliffordResponse:
        """Conjugate a Pauli element by a Clifford element."""
        return conjugate_pauli_by_clifford(request=request, client=self.quilc_client)

    def generate_randomized_benchmarking_sequence(
        self, request: RandomizedBenchmarkingRequest
    ) -> GenerateRandomizedBenchmarkingSequenceResponse:
        """Generate a randomized benchmarking sequence."""
        return generate_randomized_benchmarking_sequence(request=request, client=self.quilc_client)



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from typing import Any, Optional
from warnings import warn

from qcs_sdk import QCSClient
from qcs_sdk.compiler.quilc import QuilcClient
from qcs_sdk.qpu.translation import (
    TranslationBackend,
    get_quilt_calibrations,
    translate,
)
from qcs_sdk.qpu.translation import (
    TranslationOptions as _TranslationOptions,
)
from rpcq.messages import ParameterSpec
from typing_extensions import TypeAlias

from pyquil.api._abstract_compiler import AbstractCompiler, EncryptedProgram, QuantumExecutable
from pyquil.quantum_processor import AbstractQuantumProcessor
from pyquil.quil import Program
from pyquil.quilatom import MemoryReference
from pyquil.quilbase import Declare

QPUCompilerAPIOptions: TypeAlias = _TranslationOptions
"""
An alias of `qcs-sdk-python`'s `TranslationOptions` class.

See `TranslationOptions in qcs-sdk-python <https://github.com/rigetti/qcs-sdk-rust/blob/760df515ff9c88c1739fd69aeb00d8d38884345d/crates/python/qcs_sdk/qpu/translation.pyi#L139>`_
"""  # noqa: E501


class QPUCompilerNotRunning(Exception):
    pass


def parse_mref(val: str) -> MemoryReference:
    """Parse a memory reference from its string representation."""
    val = val.strip()
    try:
        if val[-1] == "]":
            name, offset = val.split("[")
            return MemoryReference(name, int(offset[:-1]))
        else:
            return MemoryReference(val)
    except Exception as e:
        raise ValueError(f"Unable to parse memory reference {val}.") from e


def _collect_memory_descriptors(program: Program) -> dict[str, ParameterSpec]:
    """Collect Declare instructions that are important for building the patch table.

    :return: A dictionary of variable names to specs about the declared region.
    """
    return {
        instr.name: ParameterSpec(type=instr.memory_type, length=instr.memory_size)
        for instr in program
        if isinstance(instr, Declare)
    }


class QPUCompiler(AbstractCompiler):
    """Client to communicate with the compiler and translation service."""

    api_options: Optional[QPUCompilerAPIOptions]

    def __init__(
        self,
        *,
        quantum_processor_id: str,
        quantum_processor: AbstractQuantumProcessor,
        timeout: float = 10.0,
        client_configuration: Optional[QCSClient] = None,
        api_options: Optional[QPUCompilerAPIOptions] = None,
        quilc_client: Optional[QuilcClient] = None,
    ) -> None:
        """Instantiate a new QPU compiler client.

        :param quantum_processor_id: Processor to target.
        :param quantum_processor: Quantum processor to use as compilation target.
        :param timeout: Time limit for requests, in seconds.
        :param client_configuration: Optional client configuration. If none is provided, a default one will be loaded.
        :param api_options: Options to pass to the QPU compiler API. See ``qcs-sdk-python`` for details.
        """
        super().__init__(
            quantum_processor=quantum_processor,
            timeout=timeout,
            client_configuration=client_configuration,
            quilc_client=quilc_client,
        )

        self.api_options = api_options
        self.quantum_processor_id = quantum_processor_id
        self._calibration_program: Optional[Program] = None

    def native_quil_to_executable(
        self, nq_program: Program, *, api_options: Optional[QPUCompilerAPIOptions] = None, **kwargs: Any
    ) -> QuantumExecutable:
        """Convert a native Quil program into an executable binary which can be executed by a QPU.

        If `api_options` is provided, it overrides the options set on `self`.
        """
        program = nq_program.out()
        backend = api_options.backend if api_options is not None else None
        backend = select_backend_for_quantum_processor_id(self.quantum_processor_id, backend)

        translated_program = translate(
            native_quil=program,
            num_shots=nq_program.num_shots,
            quantum_processor_id=self.quantum_processor_id,
            translation_options=api_options or self.api_options,
            client=self._client_configuration,
        )

        ro_sources = translated_program.ro_sources or {}

        return EncryptedProgram(
            program=translated_program.program,
            memory_descriptors=_collect_memory_descriptors(nq_program),
            ro_sources={parse_mref(mref): source for mref, source in ro_sources.items() or []},
        )

    def _fetch_calibration_program(self) -> Program:
        response = get_quilt_calibrations(
            quantum_processor_id=self.quantum_processor_id,
            client=self._client_configuration,
        )
        return Program(response)

    def get_calibration_program(self, force_refresh: bool = False) -> Program:
        """Get the Quil-T calibration program associated with the underlying QPU.

        This will return a cached reference of the calibration program if present.
        Otherwise (or if forcing a refresh), it will fetch and store the
        calibration program from the QCS API.

        Note that the returned program is a reference, meaning if you modify
        it directly, you will be updating the cached version, and future calls
        to this function will return your modified reference. If you want to
        update the program without impacting the cached version, you should
        make a copy of the returned program using `Program.copy()`

        A calibration program contains a number of DEFCAL, DEFWAVEFORM, and
        DEFFRAME instructions. In sum, those instructions describe how a Quil-T
        program should be translated into analog instructions for execution on
        hardware.

        :param force_refresh: Whether or not to fetch a new calibration program before returning.
        :returns: A Program object containing the calibration definitions.
        """
        if force_refresh or self._calibration_program is None:
            try:
                self._calibration_program = self._fetch_calibration_program()
            except Exception as ex:
                raise RuntimeError("Could not fetch calibrations") from ex

        return self._calibration_program

    def reset(self) -> None:
        """Reset the state of the QPUCompiler."""
        self._calibration_program = None


class QVMCompiler(AbstractCompiler):
    """Client to communicate with the compiler."""

    def __init__(
        self,
        *,
        quantum_processor: AbstractQuantumProcessor,
        timeout: float = 10.0,
        client_configuration: Optional[QCSClient] = None,
        quilc_client: Optional[QuilcClient] = None,
    ) -> None:
        """Client to communicate with compiler.

        :param quantum_processor: Quantum processor to use as compilation target.
        :param timeout: Number of seconds to wait for a response from the client.
        :param client_configuration: Optional client configuration. If none is provided, a default one will be loaded.
        """
        super().__init__(
            quantum_processor=quantum_processor,
            timeout=timeout,
            client_configuration=client_configuration,
            quilc_client=quilc_client,
        )

    def native_quil_to_executable(self, nq_program: Program, **kwargs: Any) -> QuantumExecutable:
        return nq_program

    def reset(self) -> None:
        pass


class IncompatibleBackendForQuantumProcessorIDWarning(Warning):
    pass


def select_backend_for_quantum_processor_id(
    quantum_processor_id: str, backend: Optional[TranslationBackend]
) -> TranslationBackend:
    """Check that the translation backend is supported for the quantum processor.

    If the translation backend is not supported by the quantum processor, a supported
    translation backend is returned.

    Aspen processors only support the V1 backend. Later processors support V2.

    Once Aspen-M-3 is EOL, this function can be removed as `TranslationBackend.V2`
    will be the only supported backend.
    """
    if backend is None:
        if quantum_processor_id.startswith("Aspen-"):
            return TranslationBackend.V1
        else:
            return TranslationBackend.V2

    if backend is TranslationBackend.V2 and quantum_processor_id.startswith("Aspen"):
        warn(
            "Backend V2 is not supported for Aspen processors. Backend has been changed to V1.",
            IncompatibleBackendForQuantumProcessorIDWarning,
            stacklevel=2,
        )
        return TranslationBackend.V1

    if backend is TranslationBackend.V1 and not quantum_processor_id.startswith("Aspen"):
        warn(
            "Backend V1 is only supported for Aspen processors. Backend has been changed to V2.",
            IncompatibleBackendForQuantumProcessorIDWarning,
            stacklevel=2,
        )
        return TranslationBackend.V2

    return backend



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import warnings
from collections.abc import Sequence
from typing import Any, Optional, Union

import numpy as np
from numpy.random.mtrand import RandomState

from pyquil.paulis import PauliSum, PauliTerm
from pyquil.pyqvm import AbstractQuantumSimulator
from pyquil.quilbase import Gate
from pyquil.simulation.matrices import KRAUS_OPS, P0, P1, QUANTUM_GATES
from pyquil.simulation.tools import all_bitstrings, lifted_gate, lifted_gate_matrix


def _term_expectation(wf: np.ndarray, term: PauliTerm, n_qubits: int) -> Any:
    # Computes <psi|XYZ..XXZ|psi>
    wf2 = wf
    for qubit_i, op_str in term._ops.items():
        if not isinstance(qubit_i, int):
            raise TypeError("Operation only supported for on PauliTerms with integer qubits.")
        # Re-use QUANTUM_GATES since it has X, Y, Z
        op_mat = QUANTUM_GATES[op_str]
        op_mat = lifted_gate_matrix(matrix=op_mat, qubit_inds=[qubit_i], n_qubits=n_qubits)
        wf2 = op_mat @ wf2

    if not isinstance(term.coefficient, complex):
        raise TypeError("Operation only supported for complex numbers")

    # `wf2` is XYZ..XXZ|psi>
    # hit it with a <psi| i.e. `wf.dag`
    return term.coefficient * (wf.conj().T @ wf2)


def _is_valid_quantum_state(state_matrix: np.ndarray, rtol: float = 1e-05, atol: float = 1e-08) -> bool:
    """Check if a quantum state is valid.

    That is, the matrix is Hermitian; trace one, and that the eigenvalues are non-negative.

    :param state_matrix: a D by D np.ndarray representing a quantum state
    :param rtol: The relative tolerance parameter in np.allclose and np.isclose
    :param atol: The absolute tolerance parameter in np.allclose and np.isclose
    :return: bool
    """
    hermitian = np.allclose(state_matrix, np.conjugate(state_matrix.transpose()), rtol, atol)
    if not hermitian:
        raise ValueError("The state matrix is not Hermitian.")
    trace_one = np.isclose(np.trace(state_matrix), 1, rtol, atol)
    if not trace_one:
        raise ValueError("The state matrix is not trace one.")
    evals = np.linalg.eigvals(state_matrix)
    non_neg_eigs = all([False if val < -atol else True for val in evals])
    if not non_neg_eigs:
        raise ValueError("The state matrix has negative Eigenvalues of order -" + str(atol) + ".")
    return hermitian and trace_one and non_neg_eigs


class ReferenceWavefunctionSimulator(AbstractQuantumSimulator):
    def __init__(self, n_qubits: int, rs: Optional[RandomState] = None):
        """Initialize a wavefunction simulator that prioritizes readability over performance.

        Please consider using
        :py:class:`PyQVM(..., wf_simulator_type=ReferenceWavefunctionSimulator)` rather
        than using this class directly.

        This class uses a flat state-vector of length 2^n_qubits to store wavefunction
        amplitudes. The basis is taken to be bitstrings ordered lexicographically with
        qubit 0 as the rightmost bit. This is the same as the Rigetti Lisp QVM.

        :param n_qubits: Number of qubits to simulate.
        :param rs: a RandomState (should be shared with the owning :py:class:`PyQVM`) for
            doing anything stochastic. A value of ``None`` disallows doing anything stochastic.
        """
        self.n_qubits = n_qubits
        self.rs = rs

        self.wf = np.zeros(2**n_qubits, dtype=np.complex128)
        self.wf[0] = complex(1.0, 0)

    def sample_bitstrings(self, n_samples: int) -> np.ndarray:
        """Sample bitstrings from the distribution defined by the wavefunction.

        Qubit 0 is at ``out[:, 0]``.

        :param n_samples: The number of bitstrings to sample
        :return: An array of shape (n_samples, n_qubits)
        """
        if self.rs is None:
            raise ValueError(
                "You have tried to perform a stochastic operation without setting the "
                "random state of the simulator. Might I suggest using a PyQVM object?"
            )
        probabilities = np.abs(self.wf) ** 2
        possible_bitstrings = all_bitstrings(self.n_qubits)
        inds = self.rs.choice(2**self.n_qubits, n_samples, p=probabilities)
        bitstrings = possible_bitstrings[inds, :]
        bitstrings = np.flip(bitstrings, axis=1)
        return bitstrings

    def do_gate(self, gate: Gate) -> "ReferenceWavefunctionSimulator":
        """Perform a gate.

        :return: ``self`` to support method chaining.
        """
        unitary = lifted_gate(gate=gate, n_qubits=self.n_qubits)
        self.wf = unitary.dot(self.wf)
        return self

    def do_gate_matrix(self, matrix: np.ndarray, qubits: Sequence[int]) -> "ReferenceWavefunctionSimulator":
        """Apply an arbitrary unitary; not necessarily a named gate.

        :param matrix: The unitary matrix to apply. No checks are done.
        :param qubits: The qubits to apply the unitary to.
        :return: ``self`` to support method chaining.
        """
        unitary = lifted_gate_matrix(matrix, list(qubits), n_qubits=self.n_qubits)
        self.wf = unitary.dot(self.wf)
        return self

    def do_measurement(self, qubit: int) -> int:
        """Measure a qubit, collapse the wavefunction, and return the measurement result.

        :param qubit: Index of the qubit to measure.
        :return: measured bit
        """
        if self.rs is None:
            raise ValueError(
                "You have tried to perform a stochastic operation without setting the "
                "random state of the simulator. Might I suggest using a PyQVM object?"
            )
        # lift projective measure operator to Hilbert space
        # prob(0) = <psi P0 | P0 psi> = psi* . P0* . P0 . psi
        measure_0 = lifted_gate_matrix(matrix=P0, qubit_inds=[qubit], n_qubits=self.n_qubits)
        proj_psi = measure_0 @ self.wf
        prob_zero = np.conj(proj_psi).T @ proj_psi

        # generate random number to 'roll' for measure
        if self.rs.uniform() < prob_zero:
            # decohere state using the measure_0 operator
            unitary = measure_0 @ (np.eye(2**self.n_qubits) / np.sqrt(prob_zero))
            self.wf = unitary.dot(self.wf)
            return 0
        else:  # measure one
            measure_1 = lifted_gate_matrix(matrix=P1, qubit_inds=[qubit], n_qubits=self.n_qubits)
            unitary = measure_1 @ (np.eye(2**self.n_qubits) / np.sqrt(1 - prob_zero))
            self.wf = unitary.dot(self.wf)
            return 1

    def expectation(self, operator: Union[PauliTerm, PauliSum]) -> float:
        """Compute the expectation of an operator.

        :param operator: The operator
        :return: The operator's expectation value
        """
        if not isinstance(operator, PauliSum):
            operator = PauliSum([operator])

        return sum(_term_expectation(self.wf, term, n_qubits=self.n_qubits) for term in operator)  # type: ignore

    def reset(self) -> "ReferenceWavefunctionSimulator":
        """Reset the wavefunction to the ``|000...00>`` state.

        :return: ``self`` to support method chaining.
        """
        self.wf.fill(0)
        self.wf[0] = complex(1.0, 0)
        return self

    def do_post_gate_noise(self, noise_type: str, noise_prob: float, qubits: list[int]) -> "AbstractQuantumSimulator":
        raise NotImplementedError("The reference wavefunction simulator cannot handle noise")


def zero_state_matrix(n_qubits: int) -> np.ndarray:
    """Construct a matrix corresponding to the tensor product of `n` ground states ``|0><0|``.

    :param n_qubits: The number of qubits.
    :return: The state matrix  ``|000...0><000...0|`` for `n_qubits`.
    """
    state_matrix = np.zeros((2**n_qubits, 2**n_qubits), dtype=np.complex128)
    state_matrix[0, 0] = complex(1.0, 0)
    return state_matrix


class ReferenceDensitySimulator(AbstractQuantumSimulator):
    """A density matrix simulator that prioritizes readability over performance.

    Please consider using
    :py:class:`PyQVM(..., wf_simulator_type=ReferenceDensitySimulator)` rather
    than using this class directly.

    This class uses a dense matrix of shape ``(2^n_qubits, 2^n_qubits)`` to store the
    density matrix.

    :param n_qubits: Number of qubits to simulate.
    :param rs: a RandomState (should be shared with the owning :py:class:`PyQVM`) for
        doing anything stochastic. A value of ``None`` disallows doing anything stochastic.
    """

    def __init__(self, n_qubits: int, rs: Optional[RandomState] = None):
        self.n_qubits = n_qubits
        self.rs = rs
        self.density: np.ndarray
        self.set_initial_state(zero_state_matrix(n_qubits)).reset()

    def set_initial_state(self, state_matrix: np.ndarray) -> "ReferenceDensitySimulator":
        """Update the initial state matrix that is initialized every time reset() is called.

        The default initial state of ReferenceDensitySimulator is ``|000...00>``.

        Note that the current state matrix, i.e. ``self.density`` is not affected by this
        method; you must change it directly or else call reset() after calling this method.

        To restore default state initialization behavior of ReferenceDensitySimulator pass in
        a ``state_matrix`` equal to the default initial state on `n_qubits` (i.e. ``|000...00>``)
        and then call ``reset()``. We have provided a helper function ``n_qubit_zero_state``
        in the ``_reference.py`` module to simplify this step.

        :param state_matrix: numpy.ndarray or None.
        :return: ``self`` to support method chaining.
        """
        rows, cols = state_matrix.shape
        if rows != cols:
            raise ValueError("The state matrix is not square.")
        if self.n_qubits != int(np.log2(rows)):
            raise ValueError("The state matrix is not defined on the same numbers of qubits as the QVM.")
        if _is_valid_quantum_state(state_matrix):
            self.initial_density = state_matrix
        else:
            raise ValueError(
                "The state matrix is not valid. It must be Hermitian, trace one, " "and have non-negative eigenvalues."
            )
        return self

    def sample_bitstrings(self, n_samples: int, tol_factor: float = 1e8) -> np.ndarray:
        """Sample bitstrings from the distribution defined by the wavefunction.

        Qubit 0 is at ``out[:, 0]``.

        :param n_samples: The number of bitstrings to sample
        :param tol_factor: Tolerance to set imaginary probabilities to zero, relative to
            machine epsilon.
        :return: An array of shape (n_samples, n_qubits)
        """
        if self.rs is None:
            raise ValueError(
                "You have tried to perform a stochastic operation without setting the "
                "random state of the simulator. Might I suggest using a PyQVM object?"
            )

        # for np.real_if_close the actual tolerance is (machine_eps * tol_factor),
        # where `machine_epsilon = np.finfo(float).eps`. If we use tol_factor = 1e8, then the
        # overall tolerance is \approx 2.2e-8.
        probabilities = np.real_if_close(np.diagonal(self.density), tol=tol_factor)
        # Next set negative probabilities to zero
        probabilities = np.array([0 if p < 0.0 else p for p in probabilities])
        # Ensure they sum to one
        probabilities = probabilities / np.sum(probabilities)
        possible_bitstrings = all_bitstrings(self.n_qubits)
        inds = self.rs.choice(2**self.n_qubits, n_samples, p=probabilities)
        bitstrings = possible_bitstrings[inds, :]
        bitstrings: np.ndarray = np.flip(bitstrings, axis=1)
        return bitstrings

    def do_gate(self, gate: Gate) -> "AbstractQuantumSimulator":
        """Perform a gate.

        :return: ``self`` to support method chaining.
        """
        unitary = lifted_gate(gate=gate, n_qubits=self.n_qubits)
        self.density = unitary.dot(self.density).dot(np.conj(unitary).T)
        return self

    def do_gate_matrix(self, matrix: np.ndarray, qubits: Sequence[int]) -> "AbstractQuantumSimulator":
        """Apply an arbitrary unitary; not necessarily a named gate.

        :param matrix: The unitary matrix to apply. No checks are done
        :param qubits: A list of qubits to apply the unitary to.
        :return: ``self`` to support method chaining.
        """
        unitary = lifted_gate_matrix(matrix=matrix, qubit_inds=qubits, n_qubits=self.n_qubits)
        self.density = unitary.dot(self.density).dot(np.conj(unitary).T)
        return self

    def do_measurement(self, qubit: int) -> int:
        """Measure a qubit and collapse the wavefunction.

        :return: The measurement result. A 1 or a 0.
        """
        if self.rs is None:
            raise ValueError(
                "You have tried to perform a stochastic operation without setting the "
                "random state of the simulator. Might I suggest using a PyQVM object?"
            )
        measure_0 = lifted_gate_matrix(matrix=P0, qubit_inds=[qubit], n_qubits=self.n_qubits)
        prob_zero = np.trace(measure_0 @ self.density)

        # generate random number to 'roll' for measurement
        if self.rs.uniform() < prob_zero:
            # decohere state using the measure_0 operator
            unitary = measure_0 @ (np.eye(2**self.n_qubits) / np.sqrt(prob_zero))
            self.density = unitary.dot(self.density).dot(np.conj(unitary.T))
            return 0
        else:  # measure one
            measure_1 = lifted_gate_matrix(matrix=P1, qubit_inds=[qubit], n_qubits=self.n_qubits)
            unitary = measure_1 @ (np.eye(2**self.n_qubits) / np.sqrt(1 - prob_zero))
            self.density = unitary.dot(self.density).dot(np.conj(unitary.T))
            return 1

    def expectation(self, operator: Union[PauliTerm, PauliSum]) -> complex:
        raise NotImplementedError("To implement")

    def reset(self) -> "AbstractQuantumSimulator":
        """Reset the current state of ReferenceDensitySimulator ``self.density`` to ``self.initial_density``.

        :return: ``self`` to support method chaining.
        """
        self.density = self.initial_density
        return self

    def do_post_gate_noise(self, noise_type: str, noise_prob: float, qubits: list[int]) -> "ReferenceDensitySimulator":
        kraus_ops = KRAUS_OPS[noise_type](p=noise_prob)
        if np.isclose(noise_prob, 0.0):
            warnings.warn(f"Skipping {noise_type} post-gate noise because noise_prob is close to 0", stacklevel=2)
            return self

        for q in qubits:
            new_density = np.zeros_like(self.density)
            for kraus_op in kraus_ops:
                lifted_kraus_op = lifted_gate_matrix(matrix=kraus_op, qubit_inds=[q], n_qubits=self.n_qubits)
                new_density += lifted_kraus_op.dot(self.density).dot(np.conj(lifted_kraus_op.T))
            self.density = new_density
        return self



"""Models and functions for working with quantum processors."""

__all__ = [
    "AbstractQuantumProcessor",
    "CompilerQuantumProcessor",
    "NxQuantumProcessor",
    "QCSQuantumProcessor",
    "get_qcs_quantum_processor",
]

from pyquil.quantum_processor._base import AbstractQuantumProcessor
from pyquil.quantum_processor.compiler import CompilerQuantumProcessor
from pyquil.quantum_processor.graph import NxQuantumProcessor
from pyquil.quantum_processor.qcs import QCSQuantumProcessor, get_qcs_quantum_processor
